import { ReactiveVar } from "meteor/reactive-var";
import { Template } from "meteor/templating";
import { Tracker } from "meteor/tracker";
import { escapeRegexChars } from "../helpers/strings";
import { createEntriesFromTypes } from "../keyValueTypes/constructionHelpers";
function KeyValueInputTemplateData() {
    return Template.currentData();
}
function KeyValueInputTemplate() {
    return Template.instance();
}
Template.keyValueInput.onCreated(function () {
    var _this = this;
    this.autorun(function () {
        var _a, _b;
        var data = KeyValueInputTemplateData();
        data.typeKind = (_a = data.typeKind, (_a !== null && _a !== void 0 ? _a : "multiple"));
        data.valueKind = (_b = data.valueKind, (_b !== null && _b !== void 0 ? _b : "multiple"));
    });
    this.entries = new ReactiveVar([]);
    this.autorun(function () { return createInitialEntries(_this); });
    this.partialEntry = new ReactiveVar(null);
    resetPartialEntry(this);
    this.defaultEntries = new ReactiveVar([]);
    this.autorun(function () { return createDefaultEntries(_this); });
    this.suggestions = new ReactiveVar(null);
    this.autorun(function () { return buildSuggestions(_this); });
});
Template.keyValueInput.onRendered(function () {
    this.textInput = this.$(".key-value-text-input");
    this.element = this.$(".key-value-input")[0];
});
Template.keyValueInput.helpers({
    isSingle: function () {
        return KeyValueInputTemplate().data.typeKind === "single";
    },
    defaultEntries: function () {
        return KeyValueInputTemplate().defaultEntries.get();
    },
    entries: function () {
        return KeyValueInputTemplate().entries.get();
    },
    partialEntry: function () {
        return KeyValueInputTemplate().partialEntry.get();
    },
    availableTypes: function () {
        var data = KeyValueInputTemplateData();
        return data.typeKind === "multiple" ? data.types.map(function (t) { return t.label; }).join(": / ") + ":" : "";
    },
    suggestions: function () {
        return KeyValueInputTemplate().suggestions.get();
    },
});
Template.keyValueInput.events({
    "click .key-value-input": function (event, templateInstance) {
        moveFocus(event, templateInstance, "text-input");
    },
    "input .key-value-input, focus .key-value-input": function (event, templateInstance) {
        if (templateInstance.data.valueKind === "single" && templateInstance.entries.get().length > 0) {
            templateInstance.textInput.val("");
        }
        buildSuggestions(templateInstance);
    },
    "keydown .key-value-text-input": function (event, templateInstance) {
        var key = event.key;
        var target = event.target;
        if (templateInstance.data.valueKind === "single" && templateInstance.entries.get().length > 0
            && key !== "Backspace" && key !== "Tab") {
            event.preventDefault();
            return;
        }
        if (key === ":") {
            if (addPartialEntry(templateInstance.textInput.val(), templateInstance)) {
                event.preventDefault();
                templateInstance.textInput.val("");
                buildSuggestions(templateInstance);
            }
        }
        else if (key === "Enter" || key === "Tab") {
            if (templateInstance.textInput.val() !== "" ||
                (templateInstance.partialEntry.get() && templateInstance.data.typeKind !== "single")) {
                if (key === "Enter" || templateInstance.data.valueKind !== "single") {
                    event.preventDefault();
                }
                if (completeEntry(templateInstance.textInput.val(), templateInstance)) {
                    templateInstance.textInput.val("");
                    buildSuggestions(templateInstance);
                }
            }
        }
        else if (key === "Backspace" && target.selectionStart === 0 && target.selectionEnd === 0) {
            var prependText = editLastEntry(templateInstance);
            templateInstance.textInput.val(prependText + templateInstance.textInput.val());
            target.setSelectionRange(prependText.length, prependText.length);
            buildSuggestions(templateInstance);
            event.preventDefault();
        }
        else if (key === "Escape") {
            templateInstance.suggestions.set(null);
        }
    },
    "keyup .key-value-text-input": function (event, templateInstance) {
        var _a, _b;
        var key = event.key;
        var target = event.target;
        if (key === " " && templateInstance.data.valueKind !== "single") {
            var inputValue = templateInstance.textInput.val();
            var cursorPos = target.selectionStart;
            var value = inputValue.trim();
            if (cursorPos === inputValue.length) {
                if (!value.startsWith("\"") || value.match(/^"(?:\\.|[^\\])*"$/)) {
                    if (completeEntry(value, templateInstance)) {
                        templateInstance.textInput.val("");
                        buildSuggestions(templateInstance);
                    }
                }
            }
            else if (!value.startsWith("\"")) {
                // shift cursor due to trimmed whitespace and escaped characters in front of the cursor position
                var shiftCursor = inputValue.length - inputValue.trimLeft().length + 1
                    + (_b = (_a = inputValue.substring(0, cursorPos).match(/["\\]/g)) === null || _a === void 0 ? void 0 : _a.length, (_b !== null && _b !== void 0 ? _b : 0));
                // escape \ and "
                templateInstance.textInput.val("\"" + value.replace(/\\/g, "\\\\").replace(/"/g, "\\\"") + "\"");
                target.setSelectionRange(cursorPos + shiftCursor, cursorPos + shiftCursor);
                buildSuggestions(templateInstance);
            }
        }
        else if (key === "ArrowDown" || (key === "ArrowLeft" && target.selectionEnd === 0)) {
            event.preventDefault();
            moveFocus(event, templateInstance, key);
        }
    },
    "click .key-value-suggestion": function (event, templateInstance) {
        event.preventDefault();
        if ((!templateInstance.partialEntry.get() && addPartialEntry(this.suggestion.value, templateInstance))
            || (completeEntry(this.suggestion.value, templateInstance))) {
            templateInstance.textInput.val("");
            buildSuggestions(templateInstance);
        }
    },
    "keydown .key-value-suggestion": function (event, templateInstance) {
        var key = event.key;
        var target = event.currentTarget;
        if (["ArrowUp", "ArrowDown", "Home", "End"].includes(key)) {
            event.preventDefault();
            event.stopPropagation();
            moveFocus(event, templateInstance, key);
        }
        else if (["Enter", ":", " "].includes(key)) {
            event.preventDefault();
            templateInstance.$(target).trigger("click");
        }
        else if (key.length === 1 || ["Backspace", "Delete", "ArrowLeft", "ArrowRight"].includes(key)) {
            moveFocus(event, templateInstance, "text-input");
        }
    },
    "keydown .key-value-entry": function (event, templateInstance) {
        var key = event.key;
        if (["ArrowLeft", "ArrowRight"].includes(key)) {
            moveFocus(event, templateInstance, key);
        }
    },
    "keyup .key-value-entry": function (event, templateInstance) {
        var key = event.key;
        if (key === "Backspace" || key === "Delete") {
            event.preventDefault();
            if (!this.entry.isPartial) {
                removeEntry(this.entry, templateInstance);
                moveFocus(event, templateInstance, this.index); // index should always be set for non-partial entries
            }
            else if (templateInstance.data.typeKind !== "single") { // single type always needs a partial entry
                templateInstance.partialEntry.set(null);
                moveFocus(event, templateInstance, "ArrowRight");
            }
        }
    },
    "click .key-value-entry": function (event) {
        event.stopPropagation();
        event.currentTarget.focus();
    },
    "click .key-value-entry .fa-times": function (event, templateInstance) {
        event.stopPropagation();
        removeEntry(this.entry, templateInstance);
        moveFocus(event, templateInstance, "text-input");
    },
    "click .fa-times.remove-all": function (event, templateInstance) {
        event.stopPropagation();
        removeAllEntries(templateInstance);
        moveFocus(event, templateInstance, "text-input");
    },
    "clearEntries": function (event, templateInstance) {
        removeAllEntries(templateInstance);
    },
});
function moveFocus(event, templateInstance, to) {
    // Set focus
    if (to === "text-input") {
        templateInstance.textInput[0].focus();
        return;
    }
    else if (typeof to === "number") { // entry
        Tracker.afterFlush(function () {
            var siblings = templateInstance.$(".key-value-entry, .key-value-text-input");
            if (siblings[to + templateInstance.defaultEntries.get().length]) {
                siblings[to + templateInstance.defaultEntries.get().length].focus();
            }
        });
        return;
    }
    // Move Focus
    var target;
    var currentTarget = event.currentTarget;
    if (to === "ArrowDown") { // suggestion or text input
        if (currentTarget.classList.contains("key-value-suggestion")) {
            target = templateInstance.$(currentTarget).next(".key-value-suggestion");
        }
        else {
            target = templateInstance.$(".key-value-suggestion");
        }
    }
    else if (to === "ArrowUp") { // suggestion
        target = templateInstance.$(currentTarget).prev(".key-value-suggestion");
        if (!target[0]) {
            target = templateInstance.textInput;
        }
    }
    else if (to === "End") { // suggestion
        target = templateInstance.$(".key-value-suggestion").filter(":last");
    }
    else if (to === "Home") { // suggestion
        target = templateInstance.$(".key-value-suggestion");
    }
    else if (to === "ArrowRight") { // entry
        target = templateInstance.$(currentTarget).next(".key-value-entry");
        if (!target[0]) {
            target = templateInstance.textInput;
        }
    }
    else if (to === "ArrowLeft") { // entry or text input
        if (currentTarget.classList.contains("key-value-entry")) {
            target = templateInstance.$(currentTarget).prev(".key-value-entry");
        }
        else {
            target = templateInstance.$(".key-value-entry").filter(":last");
        }
    }
    if (target[0]) {
        target[0].focus();
    }
}
function resetPartialEntry(templateInstance) {
    if (templateInstance.data.typeKind === "single") {
        templateInstance.partialEntry.set(templateInstance.data.type.createPartialEntry());
    }
    else {
        templateInstance.partialEntry.set(null);
    }
}
function addPartialEntry(val, templateInstance) {
    if (templateInstance.partialEntry.get() || templateInstance.data.typeKind === "single"
        || (templateInstance.entries.get().length > 0 && templateInstance.data.valueKind === "single")) {
        return false;
    }
    var typeName = val.trim().toLowerCase();
    for (var _i = 0, _a = templateInstance.data.types; _i < _a.length; _i++) {
        var type = _a[_i];
        var entry = type.tryCreatePartialEntry(typeName);
        if (entry) {
            templateInstance.partialEntry.set(entry);
            return true;
        }
    }
    return false;
}
function getFallbackType(templateInstance) {
    if (templateInstance.data.typeKind === "multiple" && templateInstance.data.fallbackType !== undefined) {
        return templateInstance.data.types[templateInstance.data.fallbackType];
    }
    else {
        return undefined;
    }
}
function completeEntry(val, templateInstance) {
    if (templateInstance.entries.get().length > 0 && templateInstance.data.valueKind === "single") {
        return false;
    }
    var partialEntry = templateInstance.partialEntry.get();
    var fallbackType = getFallbackType(templateInstance);
    var trimmedValue = val.trim();
    if (trimmedValue.startsWith("\"")) {
        // remove leading and possible trailing ", then replace escaped characters
        trimmedValue = trimmedValue.replace(/^"((?:\\.|[^\\])*?)"?$/g, "$1").replace(/\\(.)/g, "$1").trim();
    }
    var newEntry;
    if (partialEntry) {
        newEntry = partialEntry;
        if (!newEntry.set(trimmedValue)) {
            return false;
        }
        resetPartialEntry(templateInstance);
    }
    else {
        if (fallbackType) {
            newEntry = fallbackType.tryCreateFullEntry(trimmedValue);
        }
        if (!newEntry) {
            // if nothing worked out, try to create a partial entry from the input
            return addPartialEntry(trimmedValue, templateInstance);
        }
    }
    var entries = templateInstance.entries.get();
    if (!templateInstance.data.allowDuplicates && entries.some(function (e) { return e.equals(newEntry); })) {
        return true;
    }
    entries.push(newEntry);
    templateInstance.entries.dep.changed();
    emitChange(templateInstance);
    return true;
}
function createInitialEntries(templateInstance) {
    var data = KeyValueInputTemplateData();
    if (data.entries && data.entries.length > 1 && data.valueKind === "single") {
        throw new Error("Cannot create more than one entry for input with valueKind=single.");
    }
    templateInstance.entries.set(createEntries(data.entries, templateInstance));
}
function createDefaultEntries(templateInstance) {
    var data = KeyValueInputTemplateData();
    if (data.defaultEntries && data.defaultEntries.length > 0 && data.valueKind === "single") {
        throw new Error("Cannot create default entries for input with valueKind=single.");
    }
    templateInstance.defaultEntries.set(createEntries(data.defaultEntries, templateInstance)
        .map(function (entry) { entry.isDefault = true; return entry; }));
}
function createEntries(entries, templateInstance) {
    var types = templateInstance.data.typeKind === "single" ?
        [templateInstance.data.type] : templateInstance.data.types;
    return createEntriesFromTypes((entries !== null && entries !== void 0 ? entries : []), types);
}
function editLastEntry(templateInstance) {
    var _a;
    if (templateInstance.data.typeKind !== "single") { // cannot edit partial entry of single mode input
        var partialEntry = templateInstance.partialEntry.get();
        if (partialEntry) {
            templateInstance.partialEntry.set(null);
            return partialEntry.label;
        }
    }
    var lastEntry = templateInstance.entries.get().splice(-1, 1)[0];
    if (lastEntry) {
        templateInstance.entries.dep.changed();
        var editValue = (_a = lastEntry.editText, (_a !== null && _a !== void 0 ? _a : ""));
        lastEntry.set(undefined);
        templateInstance.partialEntry.set(lastEntry);
        emitChange(templateInstance);
        return editValue;
    }
    return "";
}
function removeEntry(entry, templateInstance) {
    var filtered = templateInstance.entries.get().filter(function (e) { return e !== entry; });
    templateInstance.entries.set(filtered);
    emitChange(templateInstance);
}
function removeAllEntries(templateInstance) {
    templateInstance.entries.set([]);
    resetPartialEntry(templateInstance);
    templateInstance.textInput.val("");
    emitChange(templateInstance);
}
function emitChange(templateInstance) {
    Tracker.flush();
    if (templateInstance.element) {
        var event_1 = new CustomEvent("keyValueEntriesChanged", { detail: templateInstance.entries.get() });
        Tracker.afterFlush(function () { return window.setTimeout(function () { return templateInstance.element.dispatchEvent(event_1); }, 100); }); // make sure keyValueInput stays responsive; apply changes a little later
    }
}
function buildSuggestions(templateInstance) {
    var entries = templateInstance.entries.get();
    if (entries.length > 0 && templateInstance.data.valueKind === "single") {
        templateInstance.suggestions.set([]);
        return;
    }
    var partialEntry = templateInstance.partialEntry.get();
    var inputValue = templateInstance.textInput ? templateInstance.textInput.val().trim() : "";
    var suggestions = null;
    if (partialEntry) {
        var entrySuggestions = partialEntry.getSuggestions(inputValue, entries, templateInstance.defaultEntries.get(), !!templateInstance.data.allowDuplicates);
        var parsed = partialEntry.parse(inputValue);
        if (parsed !== undefined) {
            entrySuggestions.push({
                display: partialEntry.type.display(parsed),
                value: partialEntry.type.editText(parsed),
                fallback: true,
            });
        }
        if (entrySuggestions && entrySuggestions.length > 0) {
            suggestions = entrySuggestions;
        }
    }
    else {
        var types = templateInstance.data.types;
        if (inputValue === "") {
            suggestions = types.map(function (t) { return ({ display: { text: t.label }, value: t.id }); });
        }
        else {
            var matchingTypes = types
                .map(function (t) { return ({ display: { text: t.label }, match: t.findAlias(inputValue), value: t.id }); })
                .filter(function (o) { return o.match; });
            if (matchingTypes.length > 0) {
                suggestions = matchingTypes;
            }
            var fallbackType = getFallbackType(templateInstance);
            if (fallbackType) {
                var fallbackSuggestion = {
                    display: { text: fallbackType.label + ": " + inputValue },
                    value: inputValue,
                    fallback: true,
                };
                if (!suggestions) {
                    suggestions = [];
                }
                suggestions.push(fallbackSuggestion);
            }
        }
        if (suggestions) {
            suggestions.header = "Mit <kbd>:</kbd> bestätigen oder aus Liste wählen.";
        }
    }
    highlightMatches(suggestions, inputValue);
    templateInstance.suggestions.set(suggestions);
}
function highlightMatches(suggestions, substring) {
    if (!suggestions) {
        return;
    }
    var escaped = escapeRegexChars(substring);
    for (var _i = 0, suggestions_1 = suggestions; _i < suggestions_1.length; _i++) {
        var suggestion = suggestions_1[_i];
        if (suggestion.match) {
            suggestion.matchHtml = suggestion.match.replace(new RegExp(escaped, "ig"), "<b>$&</b>");
        }
    }
}
